/* bfcm.h */

struct _bfcm_struct {
   // the order for fcm which specify the number of elements in the context, a typical order number should be 6
   char order;
   // the context containing recent elements used for prediction
   int *context;

   // two lookup tables used by forward compression and backward compression
   int *frtb;
   int *bltb;
   // lookup table size for each, in bytes
   int table_size;
   // the lookup table size after log function
   int tbsizelg2;

   // the location for the compressed stream on the left and the right
   int l;
   int r;

   // current element index, which is actually the last compressed element in the left compressed stream
   int i;
   int current;

   // statistics for forward compress
   int hit1;
   int miss1;

   // statistics for backward compress
   int hit2;
   int miss2;

   // data stream to be compressed
   int *datain;
   // number of elements for datain
   int nelem;

   // the compressed stream
   int *stream;
   // the length for the compressed stream
   unsigned int stream_length;
};

#define STREAM_GROWTH 1024

typedef struct _bfcm_struct bfcm;


/* do the initialization, allocate the necessary space for stream and bfcm tables */
bfcm *bfcm_init (int *datain, int length, int order);

/* generate bidirectional traversable fcm stream. */
int bfcm_compress (bfcm *bfcm1);

/* destroy everything including the data stream, lookup table and the bfcm structure itself */
void bfcm_free (bfcm *bfcm1);

/* return the total length for storing the bfcm compressed stream */
int bfcm_length (bfcm *bfcm1);

/* write the whole bfcm structure along with its content to disk */
int bfcm_commit (bfcm * bfcm1, FILE * fout);

/* read in bfcm stream from file FIN */
bfcm *bfcm_loadin (FILE *fin);

/* get the element at INDEX, store the value in VALUE */
int bfcm_get_element (bfcm *bfcm1, int index, int *value);

/* validate the forward traverse, backward traverse and the get element methods */
int bfcm_validate (bfcm *bfcm1);



