#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>

#include <assert.h>

#include "parser.h"
#include "bfcm.h"

static int parse_block (trace *trace1, int index);
static int parse_dependences (trace *trace1, trace_block *tb, int port);
static int parse_values_computed (trace *trace1, trace_block *tb);

static char trace_no_compression = 0;
static int offset1 = 0;
static int offset2 = 0;
static int offset3 = 0;
static int offset4 = 0;

static int parse_block (trace *trace1, int index)
{
   int id, num_use_port, instr_addr;
   char file_name[LINEBUF], function_name[LINEBUF];
   int line_num;
   int i;
   trace_block *tb;

   assert (index >=0 && index < trace1->num_blocks);
   FILE *fin = trace1->ftrace;

   fscanf (fin, "%d %d %x", &id, &num_use_port, &instr_addr);
   fscanf (fin, "%s %s %d", file_name, function_name, &line_num);

   tb = (trace_block *) malloc (sizeof (trace_block));
   trace1->blocks[index] = tb;

   tb->id = id;
   tb->num_use_port = num_use_port;
   tb->instr_addr = instr_addr;
   tb->line_num = line_num;

   fwrite (file_name, 1, strlen (file_name) + 1, trace1->fnames);
   fwrite (function_name, 1, strlen (function_name) + 1, trace1->fnames);

   tb->file_name_offset = offset4;
   tb->file_name_length = strlen (file_name);
   offset4 += strlen (file_name) + 1;

   tb->function_name_offset = offset4;
   tb->function_name_length = strlen (function_name);
   offset4 += strlen (function_name) + 1;

//   if (strlen (file_name) != 3)
//   printf ("[%s]strlen=%d\n", function_name, strlen (function_name));
  
//  printf ("%d %d %X", id, num_use_port, instr_addr);
//  printf (" %s %s %d\n", file_name, function_name, line_num);

   if (num_use_port > 0) {
      tb->deps = (stream *) malloc (sizeof (stream) * num_use_port * 2);
      memset (tb->deps, 0, sizeof (stream) * num_use_port * 2);
   }

   for (i=0; i<num_use_port; i++)
   {
      parse_dependences (trace1, tb, i);
   }

   parse_values_computed (trace1, tb);


}

static int parse_dependences (trace *trace1, trace_block *tb, int port)
{
   int size;
   char str_size[LINEBUF];
   int i;
   int ssize;
   FILE *ftrace = trace1->ftrace;
   FILE *fsize1 = trace1->fsize1;
   FILE *fsize2 = trace1->fsize2;
   stream *s1 = &(tb->deps[port * 2]);
   stream *s2 = &(tb->deps[port * 2 + 1]);

   fscanf (ftrace, "%s", str_size);

   if (strcmp (str_size, "SIZE"))
   {
      return -1;
   }

   fscanf (ftrace, "%d", &size);
   s1->num = size;
   s2->num = size;

   if (size==0)
      goto leave;

   int *data1 = (int *) malloc (sizeof (int) * size);
   int *data2 = (int *) malloc (sizeof (int) * size);

   for (i=0; i < size; i++)
   {
      int idx, dep_id, instance;
      fscanf (ftrace, "%d:%d %d", &idx, &dep_id, &instance);
      data1[i] = dep_id;
      data2[i] = instance;
   }

   ssize = 4 * size;

   if (trace_no_compression || size <= 10) {
      fwrite (data1, sizeof (int), size, fsize1);
      fwrite (data2, sizeof (int), size, fsize2);
      s1->offset = offset1;
      s1->length = ssize;
      s1->type = 0;

      s2->offset = offset2;
      s2->length = ssize;
      s2->type = 0;


      offset1 += ssize;
      offset2 += ssize;
   }
   else {
      bfcm *bfcm1 = bfcm_init (data1, size, BFCM_ORDER);
      bfcm_compress (bfcm1);
      assert (bfcm1->order !=0);
      
      int length1 = bfcm_length (bfcm1);
      
      if (length1 < ssize) {
         // bfcm streams are better
         bfcm_commit (bfcm1, fsize1);
         s1->offset = offset1;
         s1->length = length1;
         s1->type = 1;
         offset1 += length1;
      }
      else {
         fwrite (data1, sizeof (int), size, fsize1);
         s1->offset = offset1;
         s1->length = ssize;
         s1->type = 0;

         offset1 += ssize;
      }


      // stream2
      bfcm *bfcm2 = bfcm_init (data2, size, BFCM_ORDER);
      bfcm_compress (bfcm2);
      
      int length2 = bfcm_length (bfcm2);

      if (length2 < ssize) {
         // bfcm streams are better
         bfcm_commit (bfcm2, fsize2);
         s2->offset = offset2;
         s2->length = length2;
         s2->type = 1;
         offset2 += length2;
      }
      else {
         fwrite (data2, sizeof (int), size, fsize2);
         s2->offset = offset2;
         s2->length = ssize;
         s2->type = 0;
         offset2 += ssize;
      }
      
      bfcm_free (bfcm1);
      bfcm_free (bfcm2);
   }

   free (data1);
   free (data2);

leave:

   return 0;
}

static int parse_values_computed (trace *trace1, trace_block *tb)
{
   int values;
   char value_str[LINEBUF];
   int i;
   int svalues;
   FILE *ftrace = trace1->ftrace;
   FILE *fvalue = trace1->fvalue;
   stream *s = &(tb->values);

   fscanf (ftrace, "%s", value_str);

   if (!strcmp (value_str, "NO"))
   {
      // NO VALUES
      fscanf (ftrace, "%s", value_str);
      values = 0;
   }
   else
   {
      // VALUES %d
      fscanf (ftrace, "%d", &values);
   }

   s->num = values;

   if (values==0)
      goto leave;

   int *data1 = (int *) malloc (sizeof (int) * values);

   for (i=0; i < values; i++)
   {
      int idx, hex_value;
      fscanf (ftrace, "%d:%X", &idx, &hex_value);
      data1[i] = hex_value;
   }

   svalues = values * 4;

   if (trace_no_compression || values <= 10) {
      fwrite (data1, sizeof (int), values, fvalue);
      s->type = 0;
      s->offset = offset3;
      s->length = svalues;
      offset3 += svalues;
   }
   else
   {
      bfcm *bfcm1 = bfcm_init (data1, values, BFCM_ORDER);
      bfcm_compress (bfcm1);

      int length1 = bfcm_length (bfcm1);

      if (length1 < svalues) {
         // bfcm streams are better, write to file
         bfcm_commit (bfcm1, fvalue);
         s->type = 1;
         s->offset = offset3;
         s->length = length1;
         offset3 += length1;
      }
      else {
         fwrite (data1, sizeof (int), values, fvalue);
         s->type = 0;
         s->offset = offset3;
         s->length = svalues;
         offset3 += svalues;
      }

      bfcm_free (bfcm1);
   }

   free (data1);

leave:

  return 0;
}

static void write_block (trace *trace1, int index)
{
   fwrite (trace1->blocks[index], sizeof (trace_block), 1, trace1->fblock);
}

int parser (trace *trace1)
{
   int instruction_block;
   int i;

   fscanf (trace1->ftrace, "%d", &instruction_block);

   if (instruction_block <= 0)
      return -1;

   fwrite (&instruction_block, sizeof (int), 1, trace1->fblock);

   trace1->blocks = (trace_block **) malloc (sizeof (trace_block *) * (instruction_block));
   trace1->num_blocks = instruction_block;

   for (i = 0; i < instruction_block; i++)
   {
      parse_block (trace1, i);
   }

   for (i = 0; i < trace1->num_blocks; i++)
   {
      trace_block *tb = trace1->blocks[i];
      fwrite (tb, sizeof (trace_block), 1, trace1->fblock);
      int j;
      for (j = 0; j < tb->num_use_port; j++)
      {
         fwrite (&(tb->deps[j * 2]), sizeof (stream), 1, trace1->fblock);
         fwrite (&(tb->deps[j * 2 + 1]), sizeof (stream), 1, trace1->fblock);
      }
   }
 

   return 0;
}

void dump_block_header (trace_block *tb)
{
   printf ("%d %d 0x%x %s %s %d\n", 
      tb->id, tb->num_use_port, tb->instr_addr, 
      tb->file_name, tb->function_name, tb->line_num);
}

int stream_get_element (stream *s, int index)
{
   assert (s != NULL);
   assert (index < s->num);

   int value;

   if (s->type == 0)
   {
      int *pd = (int *) s->content;
      value = pd[index];
   }
   else
   {
      bfcm *bfcm1 = (bfcm *) s->content;
      bfcm_get_element (bfcm1, index, &value);
   }

   return value;
}



/* get the whole control dependence from the trace */
void get_cd_all (trace *trace1)
{
   int i;

   printf ("Control Dependences:\n\n");
   for (i = 0; i < trace1->num_blocks; i++)
   {
      get_instruction_cd_upon (trace1, get_addr_by_block (trace1, i));
   }
}


unsigned int get_addr_by_block (trace *trace1, int index)
{
   assert (index >= 0 && index < trace1->num_blocks);

   return trace1->blocks[index]->instr_addr;
}


list *list_init ()
{
   list *s1;
   s1 = (list *) malloc (sizeof (list));
   memset (s1, 0, sizeof (list));

   return s1;
}

void list_add (list *s, int *element)
{
   int **content = s->content;
   int num = s->num;

   if (num == 0)
      content = (int **) malloc (sizeof (int *));
   else
      content = (int **) realloc (content, sizeof (int *) * (num + 1));

   content[num] = (int *) element;
   s->content = content;
   s->num += 1;

}

/* get the control dependence for block tb, i.e. a list of pairs that current block is depended upon */
list *get_instruction_cd_upon (trace *trace1, unsigned int addr)
{
   list *list1 = list_init ();
   int block_index = get_block_by_addr (trace1, addr);
   assert (block_index >=0 && block_index < trace1->num_blocks);
   trace_block *tb = trace1->blocks[block_index];
  
   if (tb->num_use_port > 0)
   {
      stream *s1 = &(tb->deps[0]);
      stream *s2 = &(tb->deps[1]);
   
      if (s1->num != 0)  // the first port always denotes control dependence
      {
         int j;
         for (j = 0; j < s1->num; j++)
         {
            int v1 = stream_get_element (s1, j);
            int v2 = stream_get_element (s2, j);
            instr_instance *ii1 = (instr_instance *) malloc (sizeof (instr_instance));

            if (v1 == -1)
               ii1->addr = -1;
            else
               ii1->addr = get_addr_by_block (trace1, v1);

            ii1->instance_no = v2;
            list_add (list1, (int *) ii1);
         }
      }
   }

   return list1;
}


/* return the list of (instruction, instance) pairs who depends upon idx */
list *get_cd_upon_instance (trace * trace1, instr_instance *idx)
{
   int i;
   int ev1 = get_block_by_addr (trace1, idx->addr);
   int ev2 = idx->instance_no;
   list *list1 = list_init ();

   for (i = 0; i < trace1->num_blocks; i++)
   {
      trace_block *tb = trace1->blocks[i];

      if (tb->num_use_port > 0)
      {
         stream *s1 = &(tb->deps[0]);
         stream *s2 = &(tb->deps[1]);
      
         if (s1->num != 0)  // the first port always denotes control dependence
         {
            int j;
            for (j = 0; j < s1->num; j++)
            {
               int v1 = stream_get_element (s1, j);
               int v2 = stream_get_element (s2, j);
               if ((unsigned int)v1 == ev1 && v2 == ev2)
               {
                  instr_instance *ii = (instr_instance *) malloc (sizeof (instr_instance));
                  ii->addr = get_addr_by_block (trace1, i);
                  ii->instance_no = j;
//                  printf ("found: %d\t%d\n", i, j);
                  list_add (list1, (int *) ii);
                  break;
               }
            }
         }
      }
   }

   return list1;
}


instr_instance *get_instance_cd_upon (trace *trace1, instr_instance *ii1)
{
   int block_id = get_block_by_addr (trace1, ii1->addr);
   instr_instance *ii2 = NULL;

   if (block_id != -1)
   {
      trace_block *tb = trace1->blocks[block_id];
      if (tb->num_use_port > 0)
      {
         stream *s1 = &(tb->deps[0]);
         stream *s2 = &(tb->deps[1]);
      
         if (s1->num != 0 && ii1->instance_no < s1->num)  // the first port always denotes control dependence
         {
            int j = ii1->instance_no;
            ii2 = (instr_instance *) malloc (sizeof (instr_instance));
            // extrace the information from two streams
            int v1 = stream_get_element (s1, j);
            int v2 = stream_get_element (s2, j);

            ii2->addr = get_addr_by_block (trace1, v1);
            ii2->instance_no = v2;
         }
      }
   }

   return ii2;
}


/* return a list of instance pairs that instance II1 is data-depended on */
list *get_instance_dd_upon (trace *trace1, instr_instance *ii1)
{
   int block_id = get_block_by_addr (trace1, ii1->addr);
   list *list1 = list_init ();

   if (block_id != -1)
   {
      trace_block *tb = trace1->blocks[block_id];
      if (tb->num_use_port > 1)
      {
         int i;
         // the second to last ports denote data dependence
         for (i = 1; i < tb->num_use_port; i++)
         {
            stream *s1 = &(tb->deps[i * 2]);
            stream *s2 = &(tb->deps[i * 2 + 1]);
            
            if (s1->num != 0 && ii1->instance_no < s1->num)  
            {
               int j = ii1->instance_no;
               instr_instance *ii2 = (instr_instance *) malloc (sizeof (instr_instance));
               // extrace the information from two streams
               int v1 = stream_get_element (s1, j);
               int v2 = stream_get_element (s2, j);
            
               ii2->addr = get_addr_by_block (trace1, v1);
               ii2->instance_no = v2;

               list_add (list1, (int *) ii2);
            }
         }
      }
   }

   return list1;

}

/* return the value produced by instance II1 */
int get_instance_value (trace *trace1, instr_instance *ii1)
{
   int block_id = get_block_by_addr (trace1, ii1->addr);
   int value = -1;

   if (block_id != -1)
   {
      trace_block *tb = trace1->blocks[block_id];
      stream *sv = &(tb->values);
      if (ii1->instance_no >= 0 && ii1->instance_no < sv->num)
      {
         int j = ii1->instance_no;

         // extrace the information from the value streams
         value = stream_get_element (sv, j);
      }
   }

   return value;
}


int get_block_by_addr (trace *trace1, unsigned int addr)
{
   int rtn = -1;
   int i;

   for (i = 0; i < trace1->num_blocks; i++)
   {
      trace_block *tb = trace1->blocks[i];
      if (tb->instr_addr == addr)
      {
         rtn = i;
         break;
      }
   }

   return rtn;
}


/* dump the information for instance II1 */
void dump_instr_instance (trace *trace1, instr_instance *ii1)
{
   printf ("0x%x[%d]#%d", ii1->addr, get_block_by_addr (trace1, ii1->addr) + 1, ii1->instance_no);
}


static void traverse_sample (bfcm *bfcm1, int *data, int nelem)
{
   int rtn = 0;
   int value;
   int i;

   fprintf (stderr, "\ntraverse_sample validating... ");

   // we can access every element in the compressed stream
   for (i = 0; i < nelem; i++)
   {
      bfcm_get_element (bfcm1, i, &value);
      if (value != data[i])
         rtn = -1;
   }

   // or backward traverse
   for (i = 0; i < nelem; i++)
   {
      bfcm_get_element (bfcm1, i, &value);
      if (value != data[i])
         rtn = -1;
   }

   // or by random access
   for (i = 0; i < nelem / 2; i++) {
      bfcm_get_element (bfcm1, i, &value);
      if (value != data[i])
         rtn = -1;
   }

   for ( /* cont */; i >= nelem / 5; i--) {
      bfcm_get_element (bfcm1, i, &value);
      if (value != data[i])
         rtn = -1;
   }

   for ( /* cont */; i < bfcm1->nelem; i++) {
      bfcm_get_element (bfcm1, i, &value);
      if (value != data[i])
         rtn = -1;
   }

   if (rtn == 0)
      fprintf (stderr, "pass\n");
   else
      fprintf (stderr, "fail\n");

}


static void compress_single_file (char *filename)
{
   int tot;
   int nelem;
   int *data;
   char fname[256];
   FILE *fin, *fout;

   sprintf (fname, "%s", filename);
   fin = fopen (fname, "r+b");
   
   sprintf (fname, "%s->stream", filename);
   fout = fopen (fname, "w+b");

   // first read in the number of elements in the input data file
   if (fread (&nelem, sizeof (int), 1, fin) != 1 || nelem <=0)
   {
      fprintf (stderr, "Error reading the data length.\n");
      exit (-1);
   }

   // allocate the space for the data stream
   data = (int *) malloc (nelem * sizeof (int));
   tot = fread (data, sizeof (int), nelem, fin);
   if (tot != nelem)
   {
      fprintf (stderr, "Error in the stream length.\n");
      exit (-1);
   }

   // do the compression
   bfcm *bfcm1;

   bfcm1 = bfcm_init (data, nelem, 6);

   // compress the stream
   bfcm_compress (bfcm1);

   // do the validation
   bfcm_validate (bfcm1);

   // do our traverse and validate
   traverse_sample (bfcm1, data, nelem);

   // commit the compressed stream to disk
   bfcm_commit (bfcm1, fout);

   // report the compression ratio
   printf ("\nratio=%d/%d\t%f\n", (sizeof (int) * nelem), (sizeof (int)) * (bfcm1->stream_length),
      (float) nelem / bfcm1->stream_length);

   // destroy the bfcm
   bfcm_free (bfcm1);

   free (data);

   fclose (fin);
   fclose (fout);

}


int compress_trace_file (char *fn)
{
   char fnbuf[LINEBUF];
   trace *trace1;
   FILE *ftrace;
   FILE *fsize1, *fsize2, *fvalue;
   FILE *fblock;
   FILE *fnames;

   printf ("Compressing the original trace file %s ... ", fn);
   fflush (stdout);

   ftrace = fopen (fn, "r");
   if (ftrace == NULL)
   {
      fprintf (stderr, "Cannot open trace file %s.\n", fn);
      exit (-1);
   }

   trace1 = (trace *) malloc (sizeof (trace));
   trace1->ftrace = ftrace;

   strcpy (fnbuf, fn);
   strcat (fnbuf, "_size1");
   if (trace_no_compression)
      strcat (fnbuf, "_nocompression");

   fsize1 = fopen (fnbuf, "w+");

   strcpy (fnbuf, fn);
   strcat (fnbuf, "_size2");
   if (trace_no_compression)
      strcat (fnbuf, "_nocompression");
   fsize2 = fopen (fnbuf, "w+");

   strcpy (fnbuf, fn);
   strcat (fnbuf, "_value");
   if (trace_no_compression)
      strcat (fnbuf, "_nocompression");
   fvalue = fopen (fnbuf, "w+");

   strcpy (fnbuf, fn);
   strcat (fnbuf, "_blocks");
   fblock= fopen (fnbuf, "w+");   

   strcpy (fnbuf, fn);
   strcat (fnbuf, "_names");
   fnames= fopen (fnbuf, "w+");   

   trace1->fsize1 = fsize1;
   trace1->fsize2 = fsize2;
   trace1->fvalue = fvalue;
   trace1->fblock = fblock;
   trace1->fnames = fnames;

   parser (trace1);

   printf ("success\n");

   fclose (fsize1);
   fclose (fsize2);
   fclose (fvalue);
   fclose (fblock);
   fclose (fnames);

   fclose (ftrace);

}


trace *loadin_trace_file (char *fn)
{
   int num_blocks;
   int i;
   char buf[LINEBUF];
   int fdvalue;

   sprintf (buf, "%s_blocks", fn);

   printf ("Loading the compressed trace file %s ... ", buf);
   
   FILE *fp = fopen (buf, "r");

   if (fp == NULL)
   {
      fprintf (stderr, "Cannot open trace file %s.\n", buf);
      exit (-1);
   }

   trace *trace1 = (trace *) malloc (sizeof (trace));

   sprintf (buf, "%s_names");
   trace1->fnames = fopen (buf, "r");

   sprintf (buf, "%s_size1");
   trace1->fsize1 = fopen (buf, "r");

   sprintf (buf, "%s_size2");
   trace1->fsize2 = fopen (buf, "r");

   sprintf (buf, "%s_value");
   trace1->fvalue = fopen (buf, "r");

   fread (&num_blocks, sizeof (int), 1, fp);

   trace1->num_blocks = num_blocks;
   trace1->blocks = (trace_block **) malloc (sizeof (trace_block *) * (num_blocks));

   for (i=0; i < num_blocks; i++)
   {
      trace_block *tb = (trace_block *) malloc (sizeof (trace_block));
      trace1->blocks[i] = tb; 

      fread (trace1->blocks[i], sizeof (trace_block), 1, fp);

      tb->file_name = (char *) malloc (tb->file_name_length + 1);
      tb->function_name= (char *) malloc (tb->function_name_length + 1);

      fseek (trace1->fnames, tb->file_name_offset, SEEK_SET);
      fread (tb->file_name, 1, tb->file_name_length + 1, trace1->fnames);

      fseek (trace1->fnames, tb->function_name_offset, SEEK_SET);
      fread (tb->function_name, 1, tb->function_name_length + 1, trace1->fnames);

      if (tb->num_use_port > 0)
         tb->deps = (stream *) malloc (sizeof (stream) * tb->num_use_port * 2);

      int j;
      for (j = 0; j < tb->num_use_port; j++)
      {
         stream *s1 = &(tb->deps[j * 2]);
         stream *s2 = &(tb->deps[j * 2 + 1]);
         fread (s1, sizeof (stream), 1, fp);
         fread (s2, sizeof (stream), 1, fp);

         if (s1->num == 0)
            continue;

         // stream 1
         fseek (trace1->fsize1, s1->offset, SEEK_SET);

         if (s1->type == 1)
         {
            // read in as bfcm
            bfcm *bfcm1 = bfcm_loadin (trace1->fsize1);
            s1->content = (void *) bfcm1;
         }
         else
         {
            int *data = (int *) malloc (sizeof (int) * s1->num);
            fread (data, sizeof (int), s1->num, trace1->fsize1);
            s1->content = (void *) data;
         }


         // stream2
         fseek (trace1->fsize2, s2->offset, SEEK_SET);
         
         if (s2->type == 1)
         {
            // read in as bfcm
            bfcm *bfcm1 = bfcm_loadin (trace1->fsize2);
            s2->content = (void *) bfcm1;
         }
         else
         {
            int *data = (int *) malloc (sizeof (int) * s2->num);
            fread (data, sizeof (int), s2->num, trace1->fsize2);
            s2->content = (void *) data;
         }

      }

      stream *s3 = &(tb->values);
      fseek (trace1->fvalue, s3->offset, SEEK_SET);

      if (s3->num != 0 && s3->type == 1)
      {
         // read in as bfcm
         bfcm *bfcm1 = bfcm_loadin (trace1->fvalue);
         s3->content = (void *) bfcm1;
      }
      else
      {
         int *data = (int *) malloc (sizeof (int) * s3->num);
         fread (data, sizeof (int), s3->num, trace1->fvalue);
         s3->content = (void *) data;
      }

   }

   printf ("success\n");

   fclose (fp);
   fclose (trace1->fsize1);
   fclose (trace1->fsize2);
   fclose (trace1->fvalue);
   fclose (trace1->fnames);

   return trace1;

}


int parser_main (int argc, char *argv[])
{
   char fname[256];

   assert (sizeof(int) == 4);

   if ((argc == 3) && (!strncasecmp (argv[1], "-single"))) {
      /* compress a single input file and validate the compressed stream */
      /* The format for the file should be a number of integers with the first number indicating
         the number of the following. For each line, there should be a 4-byte integer.
       e.g.
         number of elements
         element1
         element2
         ...
       */
      compress_single_file (argv[2]);
   }
   else if ((argc == 3) && (!strncasecmp (argv[1], "-trace"))) {
      /* compress the trace file
      The format for the trace is available at
        http://wet.cs->ucr.edu/trace_format.html
      There are three data streams included in the trace file.
      One is the stream for the values and the other two are the stream for dependences-> */
      compress_trace_file (argv[2]);
   }
   else if ((argc == 3) && (!strncasecmp (argv[1], "-1"))) {
      /* compress the trace file
      The format for the trace is available at
        http://wet.cs->ucr.edu/trace_format.html
      There are three data streams included in the trace file.
      One is the stream for the values and the other two are the stream for dependences-> */
      loadin_trace_file (argv[2]);
   }

   else if ((argc == 3) && (!strncasecmp (argv[1], "-cd"))) {
      /* compress the trace file
      The format for the trace is available at
        http://wet.cs->ucr.edu/trace_format.html
      There are three data streams included in the trace file.
      One is the stream for the values and the other two are the stream for dependences-> */
      trace *trace1 = loadin_trace_file (argv[2]);
//      get_cd (trace1);
   }
   
   else if ((argc == 4) && (!strncasecmp (argv[1], "-trace")) && (!strncasecmp (argv[2], "-nocompression"))) {
      /* compress the trace file
      The format for the trace is available at
        http://wet.cs->ucr.edu/trace_format.html
      There are three data streams included in the trace file.
      One is the stream for the values and the other two are the stream for dependences-> */
      trace_no_compression = 1;
      compress_trace_file (argv[3]);
   }

   else {
      printf("Usage: %s <option> <filename>\n"
         "options: \n"
         "  -single\t\t<filename>\n"
         "\tThe data file should contain the integer values to be compressed."
         " The first value should be the number of the following integers->\n\n"
         "  -trace\t\t<filename>\n"
         "\tThe trace file should have the same format as described on the web page.\n\n"
         "  -trace -nocompression <filename>\n"
         "\tGenerate the streams with no compression.\n"
         "\n", argv[0]);
   }

   return 0;
}




