#define LINEBUF 1024
#define BFCM_ORDER   6

struct _list
{
   int num;
   int **content;
};
typedef struct _list list;


struct _instr_instance
{
   unsigned int addr;
   int instance_no;
};
typedef struct _instr_instance instr_instance;


struct _stream
{
   int type;
   int offset;
   int length;

   int num;

   void *content;  // actual content
};
typedef struct _stream stream;


struct _trace_block
{
   int id;
   int num_use_port;
   unsigned int instr_addr;
   int file_name_offset;
   int file_name_length;
   char *file_name;
   int function_name_offset;
   int function_name_length;
   char *function_name;
   int line_num;

   // the number of deps is num_use_port
   stream *deps;

   // one stream for values
   stream values;
};
typedef struct _trace_block trace_block;


struct _trace
{
   FILE *ftrace;
   FILE *fblock;
   FILE *fsize1;
   FILE *fsize2;
   FILE *fvalue;
   FILE *fnames;

   int num_blocks;
   trace_block **blocks;   
};
typedef struct _trace trace;


/* ============= Trace compressing and loading ============= */

/* compress the original trace file */
int compress_trace_file (char *fn);

/* load in the compress trace file from disk */
trace *loadin_trace_file (char *fn);


/* ============= Accessing the compressed trace  ============= */

/* get the control dependence for instance II1, i.e. the pair that II1 is depended upon */
instr_instance *get_instance_cd_upon (trace *trace1, instr_instance *ii1);

/* return the list of (instruction, instance) pairs who depend upon idx */
list *get_cd_upon_instance (trace * trace1, instr_instance *idx);

/* return a list of instance pairs that instance II1 is data-depended on */
list *get_instance_dd_upon (trace *trace1, instr_instance *ii1);


/* return the value produced by instance II1 */
int get_instance_value (trace *trace1, instr_instance *ii1);

/* get the control dependence for instruction ADDR, i.e. the pairs that current block is depended upon */
list *get_instruction_cd_upon (trace *trace1, unsigned int addr);

/* dump the information for instance II1 */
void dump_instr_instance (trace *trace1, instr_instance *ii1);

/* get the block index for instruction address ADDR */
int get_block_by_addr (trace *trace1, unsigned int addr);

/* get the address for the block index INDEX */
unsigned int get_addr_by_block (trace *trace1, int index);



/* ============= List ============= */
list *list_init ();
void list_add (list *s, int *element);


