#include <stdio.h>

#include "parser.h"
#include "bfcm.h"

static list *get_blocks (trace *trace1, char *func_name)
{
   int i;
   list *list1 = list_init ();

   for (i = 0; i < trace1->num_blocks; i++)
   {
      char *f1 = trace1->blocks[i]->function_name;
      if (!strcmp (f1, func_name)) 
      {
         instr_instance *ii = (instr_instance *) malloc (sizeof (instr_instance));
         ii->addr = trace1->blocks[i]->instr_addr;
         ii->instance_no = 0;
         list_add (list1, (int *) ii);
      }
   }

   return list1;
}

/* dump the control dependences for instr_instance II */
static int get_cds_for_node (trace *trace1, instr_instance *ii)
{
   list *cds;
   int i;

   // get a list of (instruction, instance) pairs that depends on current (instruction, instance)
   cds = get_cd_upon_instance (trace1, ii);
   for (i = 0; i < cds->num; i++)
   {
      instr_instance *ii1 = (instr_instance *) (cds->content[i]);
      dump_instr_instance (trace1, ii);
      printf (" --> ");
      dump_instr_instance (trace1, ii1);
      printf ("\n");
   }

   for (i = 0; i < cds->num; i++)
   {
      instr_instance *ii1 = (instr_instance *) (cds->content[i]);

      // recursively find the following instances
      get_cds_for_node (trace1, ii1);
   }


   return 0;
}


/* dump the control dependences for function FUNC_NAME */
static void get_cds_for_function (trace *trace1, char *func_name)
{
   list *s1;
   int i, j;

   printf ("The control dependences in function [%s] are as following:\n", func_name);

   // get all the blocks for the function func_name
   s1 = get_blocks (trace1, func_name);

   for (i = 0; i < s1->num; i++)
   {
      instr_instance *ii = (instr_instance *) (s1->content[i]);
      get_cds_for_node (trace1, ii);
   }

}

/* dump the values for address ADDR */
static void get_values_for_addr (trace *trace1, unsigned int addr)
{
   printf ("The values for instruction at address 0x%x:\n", addr);
   trace_block *tb = trace1->blocks[get_block_by_addr (trace1, addr)];
   int i;

   // get all the values computed at addr
   for (i = 0; i < tb->values.num; i++)
   {
      instr_instance ii1;
      ii1.addr = addr;
      ii1.instance_no = i;
      printf ("%d: %d\n", i, get_instance_value (trace1, &ii1));
   }
}


int main (int argc, char *argv[])
{
   if ((argc == 3) && (!strncasecmp (argv[1], "-compress"))) {
      /* compress the trace file
      The format for the trace is available at
        http://wet.cs->ucr.edu/trace_format.html
      There are three data streams included in the trace file.
      One is the stream for the values and the other two are the stream for dependences-> */
      compress_trace_file (argv[2]);
   }
   else if ((argc == 4) && (!strncasecmp (argv[1], "-cd")))
   {
      // load in the compressed trace
      trace *trace1 = (trace *) loadin_trace_file (argv[2]);

      get_cds_for_function (trace1, argv[3]);
   }
   else if ((argc == 4) && (!strncasecmp (argv[1], "-values")))
   {
      // load in the compressed trace
      trace *trace1 = (trace *) loadin_trace_file (argv[2]);

      get_values_for_addr (trace1, strtol (argv[3], NULL, 16));
   }
   else {
      printf("Usage: %s <option> <filename> ...\n"
         "options: \n"
         "  -compress\t<filename>\n"
         "\tThe trace file should have the same format as described on the web page. "
         "Five corressponding files will be generated after compression. Then you can "
         "remove the original trace file.\n\n"
         "  -cd\t\t<filename> <function_name>\n"
         "\tIt will show the demo for getting control dependences for the funtion in the trace.\n\n"
         "  -values\t<filename> <address>\n"
         "\tIt will show the demo for getting values for the instruction address in the trace.\n"
         "\n", argv[0]);
   }


}

